import os
import gc
import sys
import emcee
import datetime
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
from multiprocessing import get_context
from typing import List, Tuple, Union, Callable, Optional, Dict, Any
from pySODM.optimization.visualization import traceplot, autocorrelation_plot

abs_dir = os.path.dirname(__file__)

def run_EnsembleSampler(
    pos: np.ndarray,  
    max_n: int,  
    identifier: str,  
    objective_function: Callable[[Union[List, np.ndarray], ...], float],  
    objective_function_args: Optional[Tuple] = None,  
    objective_function_kwargs: Optional[Dict[str, Any]] = None,  
    moves: List[Tuple[emcee.moves.Move, float]] = [
        (emcee.moves.DEMove(), 0.5*0.5*0.9), 
        (emcee.moves.DEMove(gamma0=1.0), 0.5*0.5*0.1),
        (emcee.moves.DESnookerMove(), 0.5*0.5),
        (emcee.moves.KDEMove(bw_method='scott'), 0.25),
        (emcee.moves.StretchMove(live_dangerously=True), 0.25)
    ],  
    fig_path: Optional[str] = None,  
    samples_path: Optional[str] = None,  
    print_n: int = 10,  
    discard: int = 0,
    thin: int = 1,
    backend: Optional[str] = None,  
    processes: int = 1,  
    progress: bool = True,  
    settings_dict: Dict[str, Any] = {}  
) -> Tuple[emcee.EnsembleSampler, xr.Dataset]:

    """
    Wrapper function to setup an `emcee.EnsembleSampler` and handle all backend-related tasks.
    
    Parameters:
    -----------
        - pos: np.ndarray
            - Starting position of the Markov Chains. We recommend using `perturbate_theta()`.
        - max_n: int
            - Maximum number of iterations.
        - identifier: str
            - Identifier of the expirement.
        - objective function: callable function
            - Objective function. Recommended `log_posterior_probability`.
        - objective_function_args: tuple
            - Arguments of the objective function. If using `log_posterior_probability` as objective function, use default `None`.
        - objective_function_kwargs: dict
            - Keyworded arguments of the objective function. If using `log_posterior_probability` as objective function, use default `None`.
        - fig_path: str
            - Location where the traceplot and autocorrelation plots should be saved.
        - samples_path: str
            - Location where the `.hdf5` backend and settings `.json` should be saved.
        - print_n: int
            - Print autocorrelation and trace plots every `print_n` iterations.
        - discard: int
            - Number of iterations to remove from the beginning of the markov chains ("burn-in").
        - thin: int
            - Retain only every `thin`-th iteration.
        - processes: int
            - Number of cores to use.
        - settings_dict: dict
            - Dictionary containing calibration settings or other usefull settings for long-term storage. Appended to output `samples_xr` as attributes.
            - Valid datatypes for values: (single values) int, float, str, bool. (arrays) homogeneous list of int, float, str, bool. 1D numpy array containing int, float.

    Hyperparameters:
    ----------------
        - moves: list
            - Algorithm used for updating the coordinates of walkers in an ensemble sampler. By default, pySODM uses a shotgun approach by implementing a balanced cocktail of `emcee` moves. Consult the [emcee documentation](https://emcee.readthedocs.io/en/stable/user/moves/).
        - backend: str
            - Location of backend previous sampling (samples_path + backend). If a backend is provided, the sampler is restarted from the last iteration of the previous run. Consult the [emcee documentation](https://emcee.readthedocs.io/en/stable/user/backends/).
        - progress: bool
            - Enables the progress bar.

    Returns:
    --------

    - sampler: emcee.EnsembleSampler
        - Emcee sampler object ([see](https://emcee.readthedocs.io/en/stable/user/sampler/)).
    
    - samples_xr: xarray.Dataset
        - samples formatted in an xarray.Dataset
        - scalar parameters:
            - dimensions: ['iteration', 'chain']
            - coordinates: [samples_np.shape[0], samples_np.shape[1]]
        - n-D  parameters:
            - dimensions: ['iteration', 'chain', '{parname}_dim_0', ..., '{parname}_dim_n']
            - coordinates: [samples_np.shape[0], samples_np.shape[1], parameter_shapes[parname][0], ..., parameter_shapes[parname][n]]
    """

    # Set default fig_path/samples_path as same directory as calibration script
    if not fig_path:
        fig_path = os.getcwd() + '/'
    else:
        if fig_path[-1] != '/':
            fig_path = fig_path+'/'
        fig_path = os.path.join(os.getcwd(), fig_path)
        # If it doesn't exist make it
        if not os.path.exists(fig_path):
            os.makedirs(fig_path)
    if not samples_path:
        samples_path = os.getcwd() + '/'
    else:
        if samples_path[-1] != '/':
            samples_path = samples_path+'/'
        samples_path = os.path.join(os.getcwd(), samples_path)
        # If it doesn't exist make it
        if not os.path.exists(samples_path):
            os.makedirs(samples_path)
    # Determine current date
    run_date = str(datetime.date.today())
    # Derive nwalkers, ndim from shape of pos
    nwalkers, ndim = pos.shape
    # Append thin and discard to settings dictionary
    settings_dict.update({'discard': discard, 'thin': thin})
    # Validate the types in the settings dictionary --> otherwise error dumping to xarray
    _validate_settings_dict(settings_dict)
    # Start printout for user
    print(f'\nMarkov-Chain Monte-Carlo sampling')
    print(f'=================================\n')   
    # By default: set up a fresh hdf5 backend in samples_path
    if not backend:
        fn_backend = str(identifier)+'_BACKEND_'+run_date+'.hdf5'
        backend = emcee.backends.HDFBackend(samples_path+fn_backend)
        backend.reset(nwalkers, ndim)

        print(f"Created new backend: '{samples_path+fn_backend}'\n")
        print(f"Starting new run")
        print(f"----------------\n")
        print(f"Parameters: {ndim}")
        print(f"Markov chains: {nwalkers}")
        print(f"Cores: {processes}")
        print(f"Iterations: {max_n}")
        print(f"Automatically checking convergence every {print_n} iterations")
        print(f"Saving samples in an xarray.Dataset every {print_n} iterations")
        print(f"Printing traceplot and autocorrelation plot every {print_n} iterations")
        print(f"Samples: {samples_path+identifier+'_SAMPLES_'+run_date+'.nc'}")
        print(f"Traceplot: {fig_path+identifier+'_TRACE_'+run_date+'.pdf'}")
        print(f"Autocorrelation plot: {fig_path+identifier+'_AUTOCORR_'+run_date+'.pdf'}\n")

    # If user provides an existing backend: continue sampling 
    else:
        backend_path=os.path.join(os.getcwd(), samples_path+backend)
        try:
            backend = emcee.backends.HDFBackend(backend)
            pos = backend.get_chain(discard=0, thin=1, flat=False)[-1, ...]
        except:
            raise FileNotFoundError("backend not found.")    
        
        print(f"Found existing backend:'{backend_path}'\n")
        print(f"Continuing run")
        print(f"--------------\n")
        print(f"Parameters: {ndim}")
        print(f"Markov chains: {nwalkers}")
        print(f"Cores: {processes}")
        print(f"Iterations: {max_n} (found {backend.get_chain(discard=0, thin=1, flat=False).shape[0]} previous iterations)")
        print(f"Automatically checking convergence every {print_n} iterations")
        print(f"Saving samples in an xarray.Dataset every {print_n} iterations")
        print(f"Printing traceplot and autocorrelation plot every {print_n} iterations")
        print(f"Samples: {samples_path+identifier+'_SAMPLES_'+run_date+'.nc'}")
        print(f"Traceplot: {fig_path+identifier+'_TRACE_'+run_date+'.pdf'}")
        print(f"Autocorrelation plot: {fig_path+identifier+'_AUTOCORR_'+run_date+'.pdf'}\n")
    sys.stdout.flush()

    # This will be useful to testing convergence
    old_tau = np.inf

    with get_context("spawn").Pool(processes=processes) as pool:

        # setup sampler
        sampler = emcee.EnsembleSampler(nwalkers, ndim, objective_function, backend=backend, pool=pool,
                        args=objective_function_args, kwargs=objective_function_kwargs, moves=moves)
        
        # deduce starting iteration
        sampler_iteration_0 = sampler.iteration 

        # run sampler
        for _ in sampler.sample(pos, iterations=max_n, progress=progress, store=True, tune=False):

            # Only automatically check convergence + printouts every `print_n` steps
            if not ((sampler.iteration > 0 and sampler.iteration % print_n == 0) or (sampler_iteration_0 + sampler.iteration == max_n)):
                continue

            ###############################
            ## UPDATE DIAGNOSTIC FIGURES ##
            ###############################
            
            # Update autocorrelation plot
            _, tau = autocorrelation_plot(sampler.get_chain(), labels=objective_function.expanded_labels,
                                            filename=fig_path+identifier+'_AUTOCORR_'+run_date+'.pdf',
                                            plt_kwargs={})
            # Update traceplot
            traceplot(sampler.get_chain(),labels=objective_function.expanded_labels,
                        filename=fig_path+identifier+'_TRACE_'+run_date+'.pdf',
                        plt_kwargs={'linewidth': 1,'color': 'black','alpha': 0.2})
            # Garbage collection
            plt.close('all')
            gc.collect()

            ###################################
            ## SAVE SAMPLES IN XARRAY FORMAT ##
            ###################################

            samples_xr = _dump_sampler_to_xarray(sampler.get_chain(discard=discard, thin=thin), objective_function.parameter_shapes,
                                                    filename=samples_path+identifier+'_SAMPLES_'+run_date+'.nc', settings_dict=settings_dict)

            #######################
            ## CHECK CONVERGENCE ##
            #######################

            # Hardcode threshold values defining convergence
            thres_multi = 50.0
            thres_frac = 0.03
            # Check if chain length > 50*max autocorrelation
            converged = np.all(np.max(tau) * thres_multi < sampler.iteration)
            # Check if average tau varied more than three percent
            converged &= np.all(np.abs(np.mean(old_tau) - np.mean(tau)) / np.mean(tau) < thres_frac)
            # Update tau
            old_tau = tau
            if converged:
                print(f'Convergence: The chain is longer than 50 times the intergrated autocorrelation time.\n')
                sys.stdout.flush()
                break 
            else:
                print(f'Non-convergence: The chain is shorter than 50 times the integrated autocorrelation time.\n')
                sys.stdout.flush()

    return sampler, samples_xr


def perturbate_theta(theta: Union[List[float], np.ndarray],
                     pert: Union[List[float], np.ndarray],
                     multiplier: int=2,
                     bounds: Optional[List[Tuple[float, float]]] = None,
                     verbose: Optional[bool]=None) -> Tuple[int, int, np.ndarray]:
    """ A function to perturbate a PSO estimate and construct a matrix with initial positions for the MCMC chains

    Parameters
    ----------

    - theta : list (containing floats) or np.array
        - Parameter estimate.

    - pert : list (containing floats) or np.array
        - Relative perturbation factors (plus-minus) on parameter estimate. Must have the same length as `theta`.

    - multiplier : int
        - Multiplier determining the total number of markov chains that will be run by emcee.
        - Typically, total nr. chains = multiplier * nr. parameters
        - Default (minimum): 2 (one chain will result in an error in emcee)
        
    - bounds : array of tuples of floats
        - Ordered boundaries for the parameter values, e.g. ((0.1, 1.0), (1.0, 10.0)) if there are two parameters.
        - Note: bounds must not be zero, because the perturbation is based on a percentage of the value, and any percentage of zero returns zero, causing an error regarding linear dependence of walkers
        
    - verbose : boolean
        - Print user feedback to stdout

    Returns
    -------
    - ndim : int
        - Number of parameters

    - nwalkers : int
        - Number of chains

    - pos : np.array
        - Initial positions for markov chains. Dimensions: [nwalkers, ndim]
    """

    # Validation
    if len(theta) != len(pert):
        raise Exception('The parameter value array "theta" must have the same length as the perturbation value array "pert".')
    if len(bounds) != len(theta):
        raise Exception('If bounds is not None, it must contain a tuple for every parameter in theta')
    # Convert theta to np.array
    theta = np.array(theta)
    # Define clipping values: perturbed value must not fall outside this range
    lower_bounds = [bounds[i][0]/(1-pert[i]) for i in range(len(bounds))]
    upper_bounds = [bounds[i][1]/(1+pert[i]) for i in range(len(bounds))]
    # Start loop
    ndim = len(theta)
    nwalkers = ndim*multiplier
    cond_number=np.inf
    retry_counter=0
    while cond_number == np.inf:
        if retry_counter==0:
            theta = np.clip(theta, lower_bounds, upper_bounds)
        pos = theta + theta*pert*np.random.uniform(low=-1,high=1,size=(nwalkers,ndim))
        cond_number = np.linalg.cond(pos)
        if ((cond_number == np.inf) and verbose and (retry_counter<20)):
            print("Condition number too high, recalculating perturbations. Perhaps one or more of the bounds is zero?")
            sys.stdout.flush()
            retry_counter += 1
        elif retry_counter >= 20:
            raise Exception("Attempted 20 times to perturb parameter values but the condition number remains too large.")
    if verbose:
        print('Total number of markov chains: ' + str(nwalkers)+'\n')
        sys.stdout.flush()
    return ndim, nwalkers, pos


def _dump_sampler_to_xarray(samples_np: np.ndarray, parameter_shapes: Dict[str, Tuple], filename: str=None, settings_dict: Dict=None) -> xr.Dataset:
    """
    A function converting the raw samples from `emcee` (numpy matrix) to an xarray dataset for convenience

    Parameters
    ----------

    - samples_np: np.ndarray
        - 3D numpy array, indexed: iteration, markov chain, parameter. 
        - obtained using `sampler.get_chain()`

    - parameter_shapes: dict
        - keys: parameter name, value: parameter shape (type: tuple)

    - filename: str (optional)
        - path and filename to store samples in (default: None -- samples not saved)

    - settings_dict: dict
        - contains calibration settings retained for long-term storage. appended as metadata to the xarray output.
        - valid datatypes for values: str, Number, ndarray, number, list, tuple, bytes
    
    Returns
    -------

    - samples_xr: xarray.Dataset
        - samples formatted in an xarray.Dataset
        - scalar parameters:
            - dimensions: ['iteration', 'chain']
            - coordinates: [samples_np.shape[0], samples_np.shape[1]]
        - n-D  parameters:
            - dimensions: ['iteration', 'chain', '{parname}_dim_0', ..., '{parname}_dim_n']
            - coordinates: [samples_np.shape[0], samples_np.shape[1], parameter_shapes[parname][0], ..., parameter_shapes[parname][n]]
    """

    data = {}
    count=0
    for name,shape in parameter_shapes.items():
        # pre-allocate dims and coords
        dims = ['iteration', 'chain']
        coords = {'iteration': range(samples_np.shape[0]), 'chain': range(samples_np.shape[1])}
        # multi-dimensional parameter
        if shape != (1,):
            # samples
            param_samples = samples_np[:, :, count : count + np.prod(shape)]                            # extract relevant slice
            reshaped_samples = param_samples.reshape(samples_np.shape[0], samples_np.shape[1], *shape)  # reshape into nD array
            arr = reshaped_samples
            count += np.prod(shape)
            # dimensions and coordinates
            for i, len_dim in enumerate(shape):
                dims.append(f'{name}_dim_{i}')
                coords[f'{name}_dim_{i}'] = range(len_dim)
        # scalar parameter
        else:  # Scalar parameter case
            arr = samples_np[:, :, count]  # Keep as 2D array
            count += 1
        # wrap in an xarray
        data[name] = xr.DataArray(arr, dims=dims, coords=coords)

    # combine it all
    samples_xr = xr.Dataset(data)

    # append settings as attributes
    samples_xr.attrs.update(settings_dict)

    # save it
    if filename:
        samples_xr.to_netcdf(filename)

    return samples_xr

def _validate_settings_dict(settings_dict: Dict) -> None:
    """
    Checks if the entries in `settings_dict` are compatible for export to netCDF.
    
    Valid types are:
        - Scalar values: int, float, bool, str
        - Arrays: homogeneous lists of int, float, bool, str; 1D NumPy arrays containing int, float
    """

    def is_homogeneous_list(lst):
        """Check if list is homogeneous and contains only allowed types."""
        if not isinstance(lst, list) or not lst:
            return False
        first_type = type(lst[0])
        return all(isinstance(item, first_type) for item in lst) and first_type in {int, float, str, bool}

    def is_valid_numpy_array(arr):
        """Check if NumPy array is 1D and contains only int or float values."""
        return isinstance(arr, np.ndarray) and arr.ndim == 1 and arr.dtype.kind in {'i', 'f'}

    for key, value in settings_dict.items():
        if isinstance(value, (int, float, bool, str, list, np.ndarray)):
            if isinstance(value, list):
                if not is_homogeneous_list(value):
                    raise TypeError(f"Invalid datatype in `settings_dict` for key: '{key}' (type: {type(value)}). lists must be homogeneous and contain only int, float, bool or str.")
            elif isinstance(value, np.ndarray):
                if not is_valid_numpy_array(value):
                    raise TypeError(f"Invalid datatype in `settings_dict` for key: '{key}' (type: {type(value)}). numpy arrays must be 1D and contain only int or float.")
        else:
            raise TypeError(f"Invalid datatype in `settings_dict` for key: '{key}' (type: {type(value)}).\nvalid types are: (scalar values) int, float, bool, str. (arrays) homogeneous lists containing int, float, str, bool. 1D numpy arrays containing int or float.")
    pass


import sys
import copy
import numpy as np
import multiprocessing as mp
from functools import partial
from typing import Callable, List, Tuple, Optional, Union, Dict, Any

"""
    A pure Python/Numpy implementation of the Nelder-Mead simplex algorithm with multiprocessing support.
    Reference: https://en.wikipedia.org/wiki/Nelder%E2%80%93Mead_method
"""

def _obj_wrapper(func, args, kwargs, x):
    """
    A posterior probability must be maximised; so we switch signs
    """
    return -func(x, *args, **kwargs)

def optimize(
    func: Callable[..., float],  
    x_start: Union[List[float], np.ndarray],  
    step: Union[List[float], np.ndarray],  
    bounds: Optional[List[Tuple[float, float]]] = None,  
    args: Tuple[Any] = (),  
    kwargs: Dict[str, Any] = {},  
    processes: int = 1,  
    no_improve_thr: float = 1e-6,  
    no_improv_break: int = 100,  
    max_iter: int = 1000,  
    alpha: float = 1.0,  
    gamma: float = 2.0,  
    rho: float = -0.5,  
    sigma: float = 0.5  
    ) -> Tuple[List[float], float]:  
    
    """
    Perform a Nelder-Mead simplex optimization -- minimization of an objective function

    Parameters
    ==========

    - func : callable function or class 'log_posterior_probability' (~/src/optimization/objective_functions.py)
        - The objective function to be minimized
    - x_start: list or 1D np.ndarray
        - Starting estimate for the search algorithm. Length must equal the number of provided bounds.
    - step: list or 1D np.ndarray
        - Size of the initial search simplex   

    Optional
    ========    

    - bounds: list containing tuples
        - The bounds of the parameter(s). In the form: [(lower, upper), ..., (lower, upper)].
        - If `func` is pySODM class 'log_posterior_probability', bounds are automatically inferred.
        - If bounds are provided anyway these overwrite the bounds available in the 'log_posterior_probability' object.
    - args : tuple
        - Additional arguments passed to objective function
    - kwargs : dict
        - Additional keyword arguments passed to objective function

    Returns
    =======

    - theta: list
        - optimised parameter values
    
    - score: float
        - corresponding objective function value
    """

    ##################
    ## Input checks ##
    ##################

    if not bounds:
        try:
            bounds = func.expanded_bounds
        except:
            raise Exception(
                "please provide 'bounds'."
            )

    # Input check bounds
    lb, ub = [], []
    for variable_bounds in bounds:
        lb.append(variable_bounds[0])
        ub.append(variable_bounds[1])
    assert len(lb) == len(ub), 'Lower- and upper-bounds must be the same length'
    assert hasattr(func, '__call__'), 'Invalid function handle'
    lb = np.array(lb)
    ub = np.array(ub)
    assert np.all(ub > lb), 'All upper-bound values must be greater than lower-bound values'

    # Convert x_start to a numpy array
    if isinstance(x_start,list):
        x_start = np.array(x_start)

    # Check length of x_start
    assert len(x_start) == len(bounds), 'Length of starting estimate must be equal to the provided number of bounds'

    # Check length of step
    assert len(x_start) == len(step), "Length of 'steps' must equal the number of parameters"

    # Construct objective function wrapper
    obj = partial(_obj_wrapper, func, args, kwargs)

    #####################
    ## Initial simplex ##
    #####################

    print(f'\nNelder-Mead minimization')
    print(f'========================\n')

    print(f'For {max_iter} iterations using {processes} cores')
    print(f'Starting point: {x_start}')
    print(f'Bounds: {bounds}\n')
    sys.stdout.flush()
    
    # Compute score of initial estimate
    dim = len(x_start)
    prev_best = obj(x_start)
    no_improv = 0
    res = [[x_start, prev_best]]
    # Perturbate and construct list of arguments for objective function
    mp_args = []
    for i in range(dim):
        x = copy.copy(x_start)
        x[i] = x[i] + step[i]*x[i]
        mp_args.append(x)
    # Compute scores
    if processes > 1:
        mp_pool = mp.Pool(processes)
        score = mp_pool.map(obj, mp_args)
        mp_pool.close()
    else:
        score=[]
        for x in mp_args:
            score.append(obj(x))
    # Check bounds
    for i,x in enumerate(mp_args):
        for j, _ in enumerate(x):
            if ((x[j] < lb[j]) | (x[j] > ub[j])):
                score[i] = np.inf 
    # Construct vector of inputs and scores
    for i in range(len(score)):
        res.append([mp_args[i], score[i]])
    # Order scores
    res.sort(key=lambda x: x[1])
    best = res[0][1]

    # simplex iter
    iters = 0
    while 1:

        #######################
        ## Check convergence ##
        #######################

        # Order scores
        res.sort(key=lambda x: x[1])
        best = res[0][1]
        # Print current score
        print(f'Best after iteration {str(iters)}: score: {best:.3e}; parameters: {res[0][0]}')
        # Check if we need to stop
        # Break after max_iter
        if max_iter and iters >= max_iter:
            print('Maximum number of iteration reached. Quitting.\n')
            return res[0][0], res[0][1]
        iters += 1
        # Break if no improvements for too long
        if best < prev_best - no_improve_thr:
            no_improv = 0
            prev_best = best
        else:
            no_improv += 1
        if no_improv >= no_improv_break:
            print('Maximum number of iterations without improvement reached. Quitting.\n')
            return res[0][0], res[0][1]
        # Force printing on clusters
        sys.stdout.flush()

        ################
        ## Reflection ##
        ################

        # Construct centroid
        x0 = [0.] * dim
        for tup in res[:-1]:
            for i, c in enumerate(tup[0]):
                x0[i] += c / (len(res)-1)

        xr = x0 + alpha*(x0 - res[-1][0])
        rscore = obj(xr)
        # Check bounds
        for j, _ in enumerate(xr):
            if ((xr[j] < lb[j]) | (xr[j] > ub[j])):
                rscore = np.inf 
        if res[0][1] <= rscore < res[-2][1]:
            del res[-1]
            res.append([xr, rscore])
            continue

        ################
        ## Reflection ##
        ################

        if rscore < res[0][1]:
            xe = x0 + gamma*(x0 - res[-1][0])
            escore = obj(xe)
            # Check bounds
            for j, _ in enumerate(xe):
                if ((xe[j] < lb[j]) | (xe[j] > ub[j])):
                    escore = np.inf 
            if escore < rscore:
                del res[-1]
                res.append([xe, escore])
                continue
            else:
                del res[-1]
                res.append([xr, rscore])
                continue

        #################
        ## Contraction ##
        #################

        xc = x0 + rho*(x0 - res[-1][0])
        cscore = obj(xc)
        # Check bounds
        for j, _ in enumerate(xc):
            if ((xc[j] < lb[j]) | (xc[j] > ub[j])):
                escore = np.inf 
        if cscore < res[-1][1]:
            del res[-1]
            res.append([xc, cscore])
            continue

        ###############
        ## Reduction ##
        ###############

        x1 = res[0][0]
        nres = []
        # Construct list of arguments for objective function
        mp_args = []
        for tup in res:
            redx = x1 + sigma*(tup[0] - x1)
            mp_args.append(redx)
        # Compute
        if processes > 1:
            mp_pool = mp.Pool(processes)
            score = mp_pool.map(obj, mp_args)
            mp_pool.close()
        else:
            score=[]
            for x in mp_args:
                score.append(obj(x))
        # Check bounds
        for i,x in enumerate(mp_args):
            for j, _ in enumerate(x):
                if ((x[j] < lb[j]) | (x[j] > ub[j])):
                    score[i] = np.inf 
        # Construct nres
        for i in range(len(score)):
            nres.append([mp_args[i], score[i]])
        res = nres


import math
import matplotlib.pyplot as plt
import numpy as np
import emcee

def _apply_tick_locator(ax):
    """support function to apply default ticklocator settings"""
    ax.xaxis.set_major_locator(plt.MaxNLocator(5))
    ax.yaxis.set_major_locator(plt.MaxNLocator(4))
    return ax

def autocorrelation_plot(samples, labels=None, filename=None, plt_kwargs={}):
    """
    Make a visualization of autocorrelation of each chain

    Parameters
    ----------
    samples: np.array
        A 3-D numpy array containing the sampled parameters.
        The x-dimension must be the number of samples, the y-dimension the number of parallel chains and the z-dimension the number of sampled parameters.

    Returns
    -------
    ax: matplotlib axis object

    tau_vect[-1]: list
        Autocorrelation of last step
    """

    # Extract dimensions of sampler output
    _, _, ndim = samples.shape
    # Generate list of lables if none provided
    if not labels:
        labels = [f"$\\theta_{i}$" for i in range(ndim)]
    else:
        # input check
        if len(labels) != ndim:
            raise ValueError(
            "The length of label list is not equal to the length of the z-dimension of the samples.\n"
            "The list of label is of length: {0}. The z-dimension of the samples of length: {1}".format(len(labels), ndim)
            )

    # Compute autocorrelation/chain
    ndim = samples.shape[2]
    step_autocorr = math.ceil(samples.shape[0]/100)
    tau_vect = []
    index = 0
    for i in range(step_autocorr, samples.shape[0], step_autocorr):
        tau_vect.append(emcee.autocorr.integrated_time(samples[:i], tol = 0))
        index += 1
    n = step_autocorr * np.arange(1, index + 1)

    # Make figure
    _,ax=plt.subplots(figsize=(10,4))
    # Autocorrelation
    ax.plot(n, np.array(tau_vect), **plt_kwargs)
    ax.plot(n, n/50, "--k")
    ax.set_xlim(0, n.max())
    ax.set_xlabel("Iteration (-)")
    ax.set_ylabel("Integrated autocorrelation (-)");    
    ax.grid(False)
    ax.legend(labels)

    # Save result if desired
    if filename:
        plt.savefig(filename, dpi=200, bbox_inches='tight', orientation='portrait')

    return ax, tau_vect[-1]

def traceplot(samples, labels=None, filename=None, plt_kwargs={}):
    """
    Make a visualization of sampled parameters

    Parameters
    ----------
    samples: np.array
        A 3-D numpy array containing the sampled parameters.
        The x-dimension must be the number of samples, the y-dimension the number of parallel chains and the z-dimension the number of sampled parameters.
    labels: list
        A list containing the names of the sampled parameters. Must be the same length as the z-dimension of the samples np.array.
    plt_kwargs: dictionary
        A dictionary containing arguments for the `ax.plot()` function used to visualise the traces.

    Returns
    -------
    ax
    """
    # Extract dimensions of sampler output
    nsamples, _, ndim = samples.shape
    # Generate list of lables if none provided
    if not labels:
        labels = [f"$\\theta_{i}$" for i in range(ndim)]
    else:
        # input check
        if len(labels) != ndim:
            raise ValueError(
            "The length of label list is not equal to the length of the z-dimension of the samples.\n"
            "The list of label is of length: {0}. The z-dimension of the samples of length: {1}".format(len(labels), ndim)
            )

    # initialise figure
    _, axes = plt.subplots(nrows=len(labels), ncols=2, figsize=(8.3,(11.7/6)*len(labels)), width_ratios=[2.5,1])

    # visualise data
    axes =  np.atleast_2d(axes) # np.atleast_2d() solves error if number of parameters equal to one
    for i in range(ndim):
        ax = axes[i,:] 
        # traces
        ax[0].plot(samples[:, :, i], **plt_kwargs)
        ax[0].set_xlim(0, nsamples)
        ax[0].set_ylabel(labels[i])
        ax[0].grid(False)
        # marginal distribution
        d = np.random.choice(samples[:, :, i].flatten(), int(min(len(samples[:, :, i].flatten())/3, 1000)))
        ax[1].hist(d, color='black', density=True)
        ax[1].axvline(np.median(d), color='red', linewidth=1, linestyle='--')

    # axes labels
    axes[-1,0].set_xlabel("Iteration (-)")

    # Save result if desired
    if filename:
        plt.savefig(filename, dpi=200, bbox_inches='tight', orientation='portrait')

    return ax


import numpy as np
import pandas as pd
import xarray as xr
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from typing import List, Tuple, Union, Dict, Any
from pySODM.models.utils import list_to_dict
from pySODM.optimization.objective_functions import validate_calibrated_parameters


def add_poisson_noise(output: xr.Dataset) -> xr.Dataset:
    """A function to add poisson noise to a simulation result

    Parameters
    ----------

    output: xarray.Dataset
        Simulation output

    Returns
    -------

    output: xarray.Dataset
        Simulation output, but every value was replaced with a poisson estimate.
    """
    new_output = output.copy()
    # Loop over variables in xarray
    for varname, da in new_output.data_vars.items():
        # Replace very value with a poison draw
        values = np.random.poisson(da.values)
        new_output[varname].values = values
    return new_output


def add_gaussian_noise(output: xr.Dataset, sigma: float, relative: bool=True) -> xr.Dataset:
    """A function to add absolute gaussian noise to a simulation result

    Parameters
    ----------

    output: xarray.Dataset
        Simulation output

    sigma: float
        Standard deviation. Must be larger than or equal to 0. Equal for all datapoints.

    relative: bool
        Add noise relative to magnitude of simulation output.

    Returns
    -------

    output: xarray.Dataset
        Simulation output, but every value was replaced with a gaussian estimate.

    """
    new_output = output.copy()
    # Loop over variables in xarray
    for varname, da in new_output.data_vars.items():
        # Replace very value with a normal draw
        if relative == False:
            values = np.random.normal(da.values, sigma)
        elif relative == True:
            values = np.random.normal(da.values, np.abs(sigma*da.values))
        new_output[varname].values = values
    return new_output


def add_negative_binomial_noise(output: xr.Dataset, alpha: float) -> xr.Dataset:
    """A function to add negative binomial noise to a simulation result

    Parameters
    ----------

    output: xarray.Dataset
        Simulation output

    alpha: float
        Overdispersion factor. Must be larger than or equal to 0. Reduces to Poisson noise for alpha --> 0.

    Returns
    -------

    output: xarray.Dataset
        Simulation output, but every value was replaced with a negative binomial estimate.
    """

    new_output = output.copy()
    # Loop over variables in xarray
    for varname, da in new_output.data_vars.items():
        # Replace very value with a negative_binomial draw
        values = np.random.negative_binomial(
            1/alpha, (1/alpha)/(da.values + (1/alpha)))
        new_output[varname].values = values
    return new_output


def assign_theta(param_dict: Dict[str, Any], parameter_names: List[str], thetas: Union[List[float], np.ndarray]) -> Dict[str, Any]:
    """ A generic function to assign the output of a PSO/Nelder-Mead calibration to the model parameters dictionary

    Parameters
    ----------
    param_dict : dict
        Model parameters dictionary

    parameter_names : list (of strings)
        Names of model parameters estimated using PSO

    thetas : list (of floats)
        Result of a PSO or Nelder-Mead calibration, results must correspond to the order of the parameter names list (pars)

    Returns
    -------

    param_dict : dict
        Model parameters dictionary with values of parameters 'pars' set to the obtained PSO estimate in vector 'theta'

    """

    _, parameter_shapes = validate_calibrated_parameters(parameter_names, param_dict)
    thetas_dict = list_to_dict(np.array(thetas), parameter_shapes, retain_floats=True)
    for _, (param, value) in enumerate(thetas_dict.items()):
        param_dict.update({param: value})
    return param_dict


def variance_analysis(data: pd.Series, window_length: str, half_life: float) -> Tuple[pd.DataFrame, plt.Axes]:

    """ A function to analyze the relationship between the variance and the mean in a timeseries of data
        ================================================================================================

        1. An exponential moving average (EMA) of the data is computed.
        2. Data and EMA are binned in windows of length `window_length`.
        3. The mean of the EMA is computed in each bin.
        4. The variance of the data in relation to the mean of the EMA is computed in each bin.
        5. Several statistical models are then fitted to the relationship between the mean and variance. The statistical models are: gaussian (var = c), poisson (var = mu), quasi-poisson (var = theta*mu), negative binomial (var = mu + alpha*mu**2).

        To assure quality of the output, user must,

        1. Make sure the EMA of the data follows the data sufficiently closely.
        2. Make sure each window contains multiple datapoints (3 is a minimum).
        3. Assess how their results change when `window_length` and `halflife` are adjusted.

        Please consult section 3.1 of the Supplementary Materials in Alleman et al. 2023 Applied Mathematical Modeling for a rigorous description of the procedure.

        Parameters
        ----------

            series: pd.Series
                Timeseries of data to be analyzed. The series must have a pd.Timestamp index labeled 'date' for the time dimension.
                Additionally, this function supports the addition of one more dimension (f.i. space) using a multiindex.
                This function is not intended to study the variance of datasets containing multiple datapoints on the same date. 

            window_length: str
                Length of each window, given as a valid pd.Timedelta frequency.
                Valid options are: 'W': weekly, 'M': monthly, etc. Or multiples thereof: '2W': biweekly, etc.

            halflife: float
                Halflife of the exponential moving average.

        Output
        ------

            result: pd.Dataframe
                Contains the estimated parameter(s) and the Akaike Information Criterion (AIC) of the fitted statistical model.
                If two index levels are present (thus 'date' and 'other index level'), the result pd.Dataframe contains the result stratified per 'other index level'.

            ax: axes object
                Contains a plot of the estimated mean versus variance, togheter with the fitted statistical models. The best-fitting model is less transparent than the other models.
       """

    #################
    ## Bookkeeping ##
    #################

    # Input checks
    if 'date' not in data.index.names:
        raise ValueError(
            "Indexname 'date' not found. Make sure the time dimension index is named 'date'. Current index dimensions: {0}".format(
                data.index.names)
        )
    if len(data.index.names) > 2:
        raise ValueError(
            "The maximum number of index dimensions is two and must always include a time dimension named 'date'. Valid options are thus: 'date', or ['date', 'something else']. Current index dimensions: {0}".format(
                data.index.names)
        )
    # Relevant parameters
    if len(data.index.names) == 1:
        secundary_index = False
        secundary_index_name = None
        secundary_index_values = None
    else:
        secundary_index = True
        secundary_index_name = data.index.names[data.index.names != 'date']
        secundary_index_values = data.index.get_level_values(
            data.index.names[data.index.names != 'date'])

    ###########################################
    ## Define variance models and properties ##
    ###########################################

    def gaussian(mu, var): return var*np.ones(len(mu))
    def poisson(mu, dummy): return mu
    def quasi_poisson(mu, theta): return mu*theta
    def negative_binomial(mu, alpha): return mu + alpha*mu**2
    models = [gaussian, poisson, quasi_poisson, negative_binomial]
    n_model_pars = [1, 0, 1, 1]
    model_names = ['gaussian', 'poisson', 'quasi-poisson', 'negative binomial']

    ##########################################################
    ## Define error function for parameter estimation (SSE) ##
    ##########################################################

    def error(model_par, model, mu_data, var_data): return sum(
        (model(mu_data, model_par) - var_data)**2)

    #################################
    ## Approximate mu, var couples ##
    #################################

    # needed to generate data to calibrate our variance model to
    if not secundary_index:
        rolling_mean = data.ewm(halflife=half_life, adjust=False).mean()
        mu_data = (data.groupby(
            [pd.Grouper(freq=window_length, level='date')]).mean())
        var_data = (((data-rolling_mean) **
                    2).groupby([pd.Grouper(freq=window_length, level='date')]).mean())
    else:
        rolling_mean = data.groupby(level=secundary_index_name, group_keys=False).apply(
            lambda x: x.ewm(halflife=half_life, adjust=False).mean())
        mu_data = (data.groupby([pd.Grouper(
            freq=window_length, level='date')] + [secundary_index_values]).mean())
        var_data = (((data-rolling_mean)**2).groupby([pd.Grouper(
            freq=window_length, level='date')] + [secundary_index_values]).mean())

    # Protect calibration against nan values
    merge = pd.merge(mu_data, var_data, right_index=True,
                     left_index=True).dropna()
    mu_data = merge.iloc[:, 0]
    var_data = merge.iloc[:, 1]

    ###################################
    ## Preallocate results dataframe ##
    ###################################

    if not secundary_index:
        results = pd.DataFrame(index=model_names, columns=[
                               'theta', 'AIC'], dtype=np.float64)
    else:
        iterables = [data.index.get_level_values(
            secundary_index_name).unique(), model_names]
        index = pd.MultiIndex.from_product(
            iterables, names=[secundary_index_name, 'model'])
        results = pd.DataFrame(index=index, columns=[
                               'theta', 'AIC'], dtype=np.float64)

    ########################
    ## Perform estimation ##
    ########################

    if not secundary_index:
        for i, model in enumerate(models):
            opt = minimize(error, 0, args=(
                model, mu_data.values, var_data.values))
            results.loc[model_names[i], 'theta'] = opt['x'][0]
            n = len(mu_data.values)
            results.loc[model_names[i], 'AIC'] = n * \
                np.log(opt['fun']/n) + 2*n_model_pars[i]
    else:
        for index in secundary_index_values.unique():
            for i, model in enumerate(models):
                opt = minimize(error, 0, args=(model, mu_data.loc[slice(
                    None), index].values, var_data.loc[slice(None), index].values))
                results.loc[(index, model_names[i]), 'theta'] = opt['x'][0]
                n = len(mu_data.loc[slice(None), index].values)
                results.loc[(index, model_names[i]), 'AIC'] = n * \
                    np.log(opt['fun']/n) + 2*n_model_pars[i]

    ##########################
    ## Make diagnostic plot ##
    ##########################
    from itertools import compress
    linestyles = ['-', '-.', ':', '--']

    if not secundary_index:
        fig, ax = plt.subplots(figsize=(12, 4))
        ax.scatter(mu_data, var_data, color='black', alpha=0.5,
                   linestyle='None', facecolors='none', s=60, linewidth=2)
        mu_model = np.linspace(start=0, stop=max(mu_data))
        # Find model with lowest AIC
        best_model = list(
            compress(model_names, results['AIC'].values == min(results['AIC'].values)))[0]
        for idx, model in enumerate(models):
            if model_names[idx] == best_model:
                ax.plot(mu_model, model(
                    mu_model, results.loc[model_names[idx], 'theta']), linestyles[idx], color='black', linewidth='2')
            else:
                ax.plot(mu_model, model(
                    mu_model, results.loc[model_names[idx], 'theta']), linestyles[idx], color='black', linewidth='2', alpha=0.2)
            model_names[idx] += ' (AIC: {:.0f})'.format(
                results.loc[model_names[idx], 'AIC'])
        ax.grid(False)
        ax.set_ylabel('Estimated variance')
        ax.set_xlabel('Estimated mean')
        ax.legend(['data', ]+model_names, bbox_to_anchor=(0.05, 1),
                  loc='upper left', fontsize=12)

    else:
        # Compute figure size
        ncols = 3
        nrows = int(np.ceil(len(secundary_index_values.unique())/ncols))
        fig, ax = plt.subplots(ncols=ncols, nrows=nrows, figsize=(12, 12))
        i = 0
        j = 0
        for k, index in enumerate(secundary_index_values.unique()):
            # Determine right plot index
            if ((k % ncols == 0) & (k != 0)):
                j = 0
                i += 1
            elif k != 0:
                j += 1
            # Plot data
            ax[i, j].scatter(mu_data.loc[slice(None), index].values, var_data.loc[slice(
                None), index].values, color='black', alpha=0.5, facecolors='none', linestyle='None', s=60, linewidth=2)
            # Find best model
            best_model = list(compress(model_names, results.loc[(index, slice(
                None)), 'AIC'].values == min(results.loc[(index, slice(None)), 'AIC'].values)))[0]
            # Plot models
            mu_model = np.linspace(start=0, stop=max(
                mu_data.loc[slice(None), index].values))
            for l, model in enumerate(models):
                if model_names[l] == best_model:
                    ax[i, j].plot(mu_model, model(mu_model, results.loc[(
                        index, model_names[l]), 'theta']), linestyles[l], color='black', linewidth='2')
                else:
                    ax[i, j].plot(mu_model, model(mu_model, results.loc[(
                        index, model_names[l]), 'theta']), linestyles[l], color='black', linewidth='2', alpha=0.2)
            # Format axes
            ax[i, j].grid(False)
            # Add xlabels and ylabels
            if j == 0:
                ax[i, j].set_ylabel('Estimated variance')
            if i == nrows-1:
                ax[i, j].set_xlabel('Estimated mean')
            # Add a legend
            title = secundary_index_name + ': ' + str(index)
            ax[i, j].legend(['data', ]+model_names, bbox_to_anchor=(0.05, 1),
                            loc='upper left', fontsize=7, title=title, title_fontsize=8)

    return results, ax



import numpy as np
from datetime import timedelta

# time map used in `utils/time_to_date` and `validation/date_to_diff`
time_unit_map = {
    'us': 1,                       
    'ms': 1000,                    # us per ms
    's': 1000*1000,                # us per s
    'min': 60*1000*1000,           # ...
    'h': 60*60*1000*1000,        
    'D': 24*60*60*1000*1000,     
    'W': 7*24*60*60*1000*1000,   
}

def time_to_date(t, start_date, time_unit):
    """ Convert timestep to datetime
    """
    return start_date + timedelta(microseconds=t*time_unit_map[time_unit])

def list_to_dict(y, shape_dictionary, retain_floats=False):
    """
    A function to reconstruct model states, given a flat array of values and a dictionary with the variables name and desired shapes
    Does not retains floats.

    Parameters
    ----------

    y: np.ndarray
        A 1D numpy array

    shape_dictionary: dict
        A dictionary containing the desired names and shapes of the output dictionary
    
    retain_floats: bool
        If False, states initially given to pySODM as a float are converted in a 0D np.ndarray before being handed off to the integration function.
        This results in an apparent 'change' in datatype of model states to the user of an unstratified model. Computationally there are no consequences. Default: False.

    Returns
    -------

    y_dict: dict
        Dictionary containing the reconstructed values.
    """

    restoredArray =[]
    offset=0
    for s in shape_dictionary.values():
        n = np.prod(s)
        # Reshape changes type of floats to np.ndarray which is not desirable
        if ((n == 1) & (retain_floats==True)):
            restoredArray.append(y[offset])
        else:
            restoredArray.append(y[offset:(offset+n)].reshape(s))
        offset+=n
    return dict(zip(shape_dictionary.keys(), restoredArray))


def cut_ICF_parameters_from_parameters(parameters_names_modeldeclaration, extra_params_TDPF, extra_params_ICF, parameters):
    """
    A function to cut the initial condition function's parameters from the model parameters dictionary

    input
    -----

    parameters_names_modeldeclaration: list
        Parameter names provided by the user in the model declaration
    
    extra_params_TDPF: list
        Parameter names of the time-dependent parameter functions
    
    extra_params_ICF: list
        Parameter names of the initial condition function

    parameters: dict
        Model parameters dictionary

    output
    ------

    parameters: dict
        Model parameters dictionary, minus the parameters unique to the initial condition function
    """

    # Compute union of TDPF parameters and parameters in model declaration
    union_TDPF_integrate = set(extra_params_TDPF) | set(parameters_names_modeldeclaration)
    # Compute the difference between initial condition pars and union TDPF + model declaration pars to obtain unique ICF pars
    unique_ICF = set(extra_params_ICF) - union_TDPF_integrate 
    # Retain all but unique ICF pars
    return {key: value for key, value in parameters.items() if key in union_TDPF_integrate or key not in unique_ICF}



import copy
import math
import inspect
import numpy as np
from collections import OrderedDict
from datetime import datetime, timedelta

def check_formatting_names(states_names, parameters_names, parameters_stratified_names, dimensions_names):
    """
    A function checking the format of the 'states', 'parameters', 'stratified_parameters' and 'dimensions' provided by the user in their model class
    """

    # states_names/parameters_names
    for val, name in zip([states_names, parameters_names], ['states', 'parameters']):
        ## list?
        assert isinstance(val, list), f"'{name}' must be a list. found '{type(val)}'."
        ## containing strings
        if not all(isinstance(v, str) for v in val):
            raise TypeError(f"not all elements in '{name}' are of type str.")
    
    # dimensions
    if dimensions_names:
        # list?
        assert isinstance(dimensions_names, list), f"'dimensions' must be a list. found '{type(dimensions_names)}'."
        # containing only strings
        if not all(isinstance(v, str) for v in dimensions_names):
            raise TypeError(f"not all elements in 'dimensions' are of type str.")
        # exract number of dimensions
        n_dims = len(dimensions_names)

    # parameters_stratified_names
    if parameters_stratified_names:
        if not dimensions_names:
            raise TypeError(f"a model without dimensions cannot have stratified parameters.")
        elif n_dims == 1:
            # list?
            assert isinstance(parameters_stratified_names, list), f"'stratified_parameters' must be a list. found '{type(parameters_stratified_names)}'."
            # containing only strings?
            if not all(isinstance(v, str) for v in parameters_stratified_names):
                raise TypeError(f"not all elements in 'stratified_parameters' are of type str.")
        elif n_dims >= 2:
            # list?
            assert isinstance(parameters_stratified_names, list), f"'stratified_parameters' must be a list. found '{type(parameters_stratified_names)}'."
            # containing n_dims sublists?
            assert len(parameters_stratified_names) == n_dims, f"'stratified_parameters' must be a list containing {n_dims} sublists."
            if not all(isinstance(sublist, list) for sublist in parameters_stratified_names):
                raise TypeError(f"'stratified_parameters' must be a list containing {n_dims} sublists.")
            # each containing only strings OR being empty?
            for sublist in parameters_stratified_names:
                if ((not all(isinstance(v, str) for v in sublist)) & (sublist != [])):
                    raise TypeError(f"'stratified_parameters' must be a list containing {n_dims} sublists. each sublist must either be empty or contain only str.")


from pySODM.models.utils import time_unit_map
def date_to_diff(start_date, end_date, time_unit):
    """
    Computes the amount of `time_unit` steps between `start_date` and `end_date`
    Rounds upward, f.i. there are approx. 4.5 weeks in a month, but this will be rounded to 5 weeks of simulation if `time_unit` is set to weeks
    """

    # count number of days
    delta_us = (end_date - start_date).total_seconds() * 1e6
    
    # check time_unit
    if time_unit not in time_unit_map:
        raise ValueError(f"Invalid time unit '{time_unit}'. Choose from {list(time_unit_map.keys())}")
    
    # compute number of time units
    num_units = math.ceil(delta_us / time_unit_map[time_unit])

    # generate the datetime range for the output
    date_range = [start_date + timedelta(microseconds=i*time_unit_map[time_unit]) for i in range(num_units+1)]

    return [0, len(date_range)-1], date_range


def validate_simulation_time(time, time_unit):
    """Validates the simulation time of the sim() function. Various input types are converted to: time = [start_float, stop_float]"""

    date_range=[None,]
    if isinstance(time, float):
        time = [0, round(time)]
    elif isinstance(time, int):
        time = [0, time]
    elif isinstance(time, list):
        if not len(time) == 2:
            raise ValueError(f"wrong length of list-like simulation start and stop (length: {len(time)}). correct format: time=[start, stop] (length: 2).")
        else:
            # If they are all int or flat (or commonly occuring np.int/np.float)
            if all([isinstance(item, (int,float,np.int32,np.float32,np.int64,np.float64)) for item in time]):
                time = [round(item) for item in time]
            # If they are all datetime
            elif all([isinstance(item, datetime) for item in time]):
                check_chronology(time)
                time, date_range = date_to_diff(time[0], time[1], time_unit)
            # If they are all strings: assume format is YYYY-MM-DD and convert to a datetime
            elif all([isinstance(item, str) for item in time]):
                time = [datetime.strptime(item,"%Y-%m-%d") for item in time]
                check_chronology(time)
                time, date_range = date_to_diff(time[0], time[1], time_unit)
            else:
                types = [type(t) for t in time]
                raise ValueError(
                    "simulation start and stop must have the format: time=[start, stop]."
                    " 'start' and 'stop' must have the same datatype: int/float, str ('yyyy-mm-dd'), or datetime."
                    f" mixing of types is not allowed. you supplied: {types} "
                    )
    else:
        raise TypeError(
                "'time' must be 1) a single int/float representing the end of the simulation, 2) a list of format: time=[start, stop]."
            )

    # check chronology (for all cases we haven't checked yet)
    check_chronology(time)

    return time, date_range


def check_chronology(time):
    """ Check the simulation start and end chronology. Works on datetimes as well as int/float.
    """
    if time[1] < time[0]:
        raise ValueError(
            "start of simulation is chronologically after end of simulation"
        )
    elif time[0] == time[1]:
        raise ValueError(
            "start of simulation is the same as the end of simulation"
        )
    pass


def validate_solution_methods_ODE(rtol, method, tau):
    """
    Validates the input arguments of the ODE.sim() function

    input
    -----

    rtol: float
        Relative solver tolerance

    method: str
        Solver method: 'RK23', 'RK45', 'DOP853', 'Radau', 'BDF', 'LSODA'

    tau: int/float
        Discrete integration size of timestep. 
    """

    if not isinstance(rtol, float):
        raise TypeError(
            "relative solver tolerance 'rtol' must be of type float"
            )
    if not isinstance(method, str):
        raise TypeError(
            "solver method 'method' must be of type string"
            )
    if method not in ['RK23', 'RK45', 'DOP853', 'Radau', 'BDF', 'LSODA']:
        raise ValueError(
            f"invalid solution method '{method}'. valid methods: 'RK23', 'RK45', 'DOP853', 'Radau', 'BDF', 'LSODA'"
        )
    if tau != None:
        if not isinstance(tau, (int,float)):
            raise TypeError(
                "discrete timestep 'tau' must be of type int or float"
            )

def validate_solution_methods_JumpProcess(method, tau):
    """
    Validates the input arguments of the JumpProcess.sim() function

    method: str
        Solver method: 'SSA' or 'tau_leap'

    tau: int/float
        If method == 'tau_leap' --> leap size
    """

    # Input checks on solution method and timestep
    if not isinstance(method, str):
        raise TypeError(
            "solver method 'method' must be of type string"
            )
    if method not in ['tau_leap', 'SSA']:
        raise ValueError(
            f"invalid solution method '{method}'. valid methods: 'SSA' and 'tau_leap'"
        )
    if not isinstance(tau, (int,float)):
        raise TypeError(
            "discrete timestep 'tau' must be of type int or float"
            )
    
def validate_draw_function(draw_function, draw_function_kwargs, parameters, initial_states_function, state_shapes, initial_states_function_args):
    """
    Validates the draw function's input and output. Used in the sim() functions of the ODE and JumpProcess classes (base.py).
    
    Makes a call to `validate_initial_states` if an initial condition function is used

    input
    -----

    draw_function: function
        a function altering parameters in the model parameters dictionary between consecutive simulations
        has the dictionary of model parameters ('parameters') as its first obligatory input, followed by a variable number of inputs

    draw_function_kwargs: dict
        a dictionary containing the aforementioned additional inputs to the 'draw_function'
    
    parameters: dict
        the dictionary of model parameters

    initial_states_function: callable or None
        a function returning a dictionary of initial states

    state_shapes: dict
        contains the shape of every model state
    
    initial_states_function_args: list
        contains the names of the initial condition functions arguments
    """

    # check that the draw_function is a function
    if not callable(draw_function):
        raise TypeError(
            f"a 'draw function' must be callable (a function)"
        )
    # check that it's first argument is named 'parameters'
    args = list(inspect.signature(draw_function).parameters.keys())
    if args[0] != "parameters":
        raise ValueError(
            f"your draw function '{draw_function.__name__}' must have 'parameters' as its first input. Its current inputs are: '{args}'"
    )
    # check that `draw_function_kwargs` is a 'dict'
    if not isinstance(draw_function_kwargs, dict):
        raise TypeError(
            f"your `draw_function_kwargs` must be of type 'dict' but are of type '{type(draw_function_kwargs)}'"
        )
    # if draw_functions_kwargs is an empty dict and draw_function has additional kwargs the user has most likely forgotten to pass draw_function_kwargs to the sim() function
    if ((len(args[1:]) > 0) & (len(list(draw_function_kwargs.keys())) == 0)):
        raise ValueError(
            f"the draw function '{draw_function.__name__}' has {len(args[1:])} arguments in addition to the mandatory 'parameters' argument\n"
            f"have you forgotten to pass `draw_function_kwargs` to the sim() function?"
        )
    # check that it's keys have the same name as the inputs of draw_function that follow `parameters`
    if set(args[1:]) != set(list(draw_function_kwargs.keys())):
        raise ValueError(
            f"incorrect arguments passed to draw function '{draw_function.__name__}'\n"
            "keys missing in 'draw_function_kwargs': {0}. redundant keys: {1}".format(set(args[1:]).difference(list(draw_function_kwargs.keys())), set(list(draw_function_kwargs.keys())).difference(set(args[1:])))
        )
    
    # call draw function and check its outputs
    output = draw_function(copy.deepcopy(parameters), **draw_function_kwargs)
    # check if it returns one output
    if not isinstance(output, dict):
        raise TypeError(f"a draw function must return one dictionary containing the model's parameters. found type '{type(output)}'")
    # verify keys are the same on input/output 'parameters'
    if set(output.keys()) != set(parameters.keys()):
        raise ValueError(
            f"a draw function must return the dictionary of model parameters"
            f"keys in output dictionary of draw function '{draw_function.__name__}' must match the keys in input dictionary 'parameters'.\n"
            "keys missing in draw function output: {0}. redundant keys: {1}".format(set(parameters.keys()).difference(set(output.keys())), set(output.keys()).difference(set(parameters.keys())))
        )
    # if an initial condition function is used:
    if initial_states_function:
        # call it with the updated parameters and verify user didn't do anything stupid compromising the IC
        initial_states = initial_states_function(**{key: output[key] for key in initial_states_function_args})
        # verify the initial states sizes
        try:
            _ = validate_initial_states(initial_states, state_shapes)
        except Exception as e:
            error_message = f"draw function --> drawn parameter dictionary --> initial condition function --> invalid initial state.\nfound error: {str(e)}"
            raise RuntimeError(error_message) from e

def fill_initial_state_with_zero(state_names, initial_states):
    """ A function filling the undefined initial states with zeros """
    for state in state_names:
        if state in initial_states:
            state_values = initial_states[state]
    return state_values

def validate_dimensions(dimension_names, coordinates):
    """Checks if the combination of the `dimension_names` in the model defenition is compatible with the `coordinates` provided when initializing the model. Returns the dimension size of every dimension.
    """
    # Validate dimension
    if dimension_names:
        if not coordinates:
            raise ValueError(
                "dimension name provided in integrate function but no coordinates were given when model was initialised"
            )
        else:
            if set(dimension_names) != set(coordinates.keys()):
                raise ValueError(
                    "`dimension_names` do not match coordinates dictionary keys.\n"
                    "Missing dimension names: {0}. Redundant coordinates: {1}".format(set(dimension_names).difference(set(coordinates.keys())), set(coordinates.keys()).difference(set(dimension_names)))
                )
            else:
                dimension_size = []
                for i, (key,value) in enumerate(coordinates.items()):
                    try:
                        dimension_size.append(len(value))
                    except:
                        raise ValueError(
                            f"Unable to deduce dimension length from '{value}' of coordinate '{key}'"
                        )
    else:
        if coordinates:
            raise ValueError(
                f"You have provided a dictionary of coordinates with keys: {list(coordinates.keys())} upon model initialization but no variable 'dimension names' was found in the integrate function.\nDefine a variable 'dimension_names = {list(coordinates.keys())}' in the model definition."
            )
        else:
            dimension_size = [1]

    return dimension_size

def validate_dimensions_per_state(dimensions_per_state, coordinates, state_names):
    """Valide if length of `dimensions_per_state` is equal to the length of `state_names`. Check if the dimensions provided for every model state are existing dimensions."""

    # Length equal to state_names?   
    if len(dimensions_per_state) != len(state_names):
        raise ValueError(
            f"The length of `dimensions_per_state` ({len(dimensions_per_state)}) must match the length of `state_names` ({len(state_names)})"
        )
    # Contains only valid coordinates?
    for i,state_name in enumerate(state_names):
        if not all(x in coordinates.keys() for x in dimensions_per_state[i]):
            raise ValueError(
                f"The dimension names of model state '{state_name}', specified in position {i} of `dimensions_per_state` contains invalid coordinate names. Redundant names: {set(dimensions_per_state[i]).difference(set(coordinates.keys()))}"
        )

def build_state_sizes_dimensions(coordinates, state_names, dimensions_per_state):
    """A function returning three dictionaries: A dictionary containing, for every model state: 1) Its shape, 2) its dimensions, 3) its coordinates.
    """

    if not dimensions_per_state:
        if not coordinates:
            return dict(zip(state_names, len(state_names)*[(1,),] )), dict(zip(state_names, len(state_names)*[[],] )), dict(zip(state_names, len(state_names)*[[],] ))
        else:
            shape=[]
            dimension=[]
            coordinate=[]
            for key,value in coordinates.items():
                try:
                    shape.append(len(value))
                    dimension.append(key)
                    coordinate.append(value)
                except:
                    raise ValueError(
                            f"Unable to deduce dimension length from '{value}' of coordinate '{key}'"
                        )
            return dict(zip(state_names, len(state_names)*[tuple(shape),] )), dict(zip(state_names, len(state_names)*[dimension,])), dict(zip(state_names, len(state_names)*[coordinate,]))
    else:
        if not coordinates:
            raise ValueError(
                "`dimensions_per_state` found in the model defenition, however you have not provided `coordinates` when initializing the model. "
            )
        else:
            shapes=[]
            coords=[]
            for i,state_name in enumerate(state_names):
                dimensions = dimensions_per_state[i]
                if not dimensions:
                    shapes.append( (1,) )
                    coords.append( [] )
                else:
                    shape=[]
                    coord=[]
                    for dimension in dimensions:
                        try:
                            shape.append(len(coordinates[dimension]))
                            coord.append(coordinates[dimension])
                        except:
                            raise ValueError(
                                 f"Unable to deduce dimension length from '{coordinates[dimension]}' of coordinate '{dimension}'"
                            )
                    shapes.append(tuple(shape))
                    coords.append(coord)
            return dict(zip(state_names, shapes)), dict(zip(state_names, dimensions_per_state)),  dict(zip(state_names, coords))

def validate_parameter_function(func):
    # Validate the function passed to time_dependent_parameters
    sig = inspect.signature(func)
    keywords = list(sig.parameters.keys())
    if keywords[0] != "t":
        raise ValueError(
            "The first parameter of the parameter function should be 't'"
        )
    if keywords[1] != "states":
        raise ValueError(
            "The second parameter of the parameter function should be 'states'"
        )
    if keywords[2] != "param":
        raise ValueError(
            "The second parameter of the parameter function should be 'param'"
        )
    else:
        return keywords[3:]

def validate_time_dependent_parameters(parameter_names, parameter_stratified_names,time_dependent_parameters):

    extra_params = []
    all_param_names = parameter_names.copy()

    if parameter_stratified_names:
        if not isinstance(parameter_stratified_names[0], list):
            all_param_names += parameter_stratified_names
        else:
            for lst in parameter_stratified_names:
                all_param_names.extend(lst)
            
    for param, func in time_dependent_parameters.items():
        if param not in all_param_names:
            raise ValueError(
                "The specified time-dependent parameter '{0}' is not an "
                "existing model parameter".format(param))
        kwds = validate_parameter_function(func)
        extra_params.append(kwds)

    return extra_params

def get_initial_states_fuction_parameters(initial_states):
    """
    A function checking the type of the initial condition. If it's a function, returns the arguments of the function. If it's a dictionary, returns an empty list.

    input
    -----

    initial_states: dict or callable
        Dictionary containing the model's initial states. Keys: model states. Values: corresponding initial values.
        Or a function generating such dictionary.
    """

    # check if type is right
    if ((not isinstance(initial_states, dict)) & (not callable(initial_states))):
        raise TypeError(f"initial states must be: 1) a dictionary, 2) a callable function generating a dictionary. found '{type(initial_states)}'")
    
    # if it's a callable function, call it and check if it returns a dictionary
    if callable(initial_states):
        # get arguments
        args = list(inspect.signature(initial_states).parameters.keys())
    else:
        args = []

    return args

def validate_initial_states(initial_states, state_shapes):
    """
    A function to check the types and sizes of the model's initial states provided by the user.
    Automatically assumes non-specified states are equal to zero.

    input
    -----

    initial_states: dict
        Dictionary containing the model's initial states. Keys: model states. Values: corresponding initial values.

    state_shapes: dict
        Contains the shape of every model state.

    output
    ------
    
    initial_states: dict
        Dictionary containing the model's validated initial states.
        Types/Size checked, redundant initial states checked, states sorted according to `state_names`.
    """

    # check if type is right (redundant; already checked in `get_initial_states_fuction_parameters`)
    if not isinstance(initial_states, dict):
        raise TypeError("initial states should be a dictionary by this point. contact Tijs Alleman if this error occurs.")

    # validate the states shape; if not present initialise with zeros
    for state_name, state_shape in state_shapes.items():
        if state_name in initial_states:
            # if present, verify the length
            initial_states[state_name] = check_initial_states_shape(
                                        initial_states[state_name], state_shape, state_name, "initial state",
                                        )
        else:
            # Fill with zeros
            initial_states[state_name] = np.zeros(state_shape, dtype=np.float64)

    # validate all states are present (using `set` to ignore order)
    if set(initial_states.keys()) != set(state_shapes.keys()):
        raise ValueError(
            f"The specified initial states don't exactly match the predefined states. Redundant states: {set(initial_states.keys()).difference(set(state_shapes.keys()))}"
        )

    # sort the initial states to match the state_names
    initial_states = {state: initial_states[state] for state in state_shapes}

    return initial_states

def evaluate_initial_condition_function(initial_states, parameters):
    """
    A function to check the user's input type of the initial states (dict or callable).
    If callable, evaluates the user's initial state function to obtain an initial states dictionary.

    input
    -----

    initial_states: dict
        Dictionary containing the model's initial states. Keys: model states. Values: corresponding initial values.

    parameters: dict
        Contains the model's parameters.

    output
    ------

    initial_states: dict
        Dictionary containing the model's validated initial states.
        Types/Size checked, redundant initial states checked, states sorted according to `state_names`.

    initial_states_function: callable
        A function generating a valid initial state
        If 'initial_states' was already a callable: equal to 'initial_states'
        If 'initial_states' was a dictionary: a dummy function returning that dictionary
    
    initial_states_function_args: list
        Contains the arguments of the initial_states_function.
    """
    
    # check if type is right (redundant; already checked in `get_initial_states_fuction_parameters`)
    if ((not isinstance(initial_states, dict)) & (not callable(initial_states))):
        raise TypeError("initial states must be: 1) a dictionary, 2) a callable function generating a dictionary")
    
    # if it's a callable function, call it and check if it returns a dictionary
    initial_states_is_function = False
    if callable(initial_states):
        initial_states_is_function = True
        # get arguments
        initial_states_function_args = list(inspect.signature(initial_states).parameters.keys())
        # save function
        initial_states_function = initial_states
        # construct initial states
        initial_states = initial_states_function(**{key: parameters[key] for key in initial_states_function_args})

    # define a dummy initial_states_function
    if not initial_states_is_function:
        initial_states_function = None
        initial_states_function_args = []

    return initial_states, initial_states_function, initial_states_function_args

def check_initial_states_shape(values, desired_shape, name, object_name):
    """ A function checking if the provided initial states have the correct shape
        Converts all values of initial states to type np.float64
    """

    # If the model doesn't have dimensions, initial states can be defined as: np.array([int/float]), [int/float], int or float
    # However these still need to converted to a np.array internally
    if list(desired_shape) == [1]:
        if not isinstance(values, (list,int,float,np.int32,np.int64,np.float32,np.float64,np.ndarray)):
            raise TypeError(
                f"{object_name} {name} must be of type int, float, or list. found {type(values)}"
            )
        else:
            if isinstance(values,(int,float)):
                values = np.asarray([values,])
        values = np.asarray(values, dtype=np.float64)

        if values.shape != desired_shape:
            raise ValueError(
                "The desired shape of model state '{name}' is {desired_shape}, but provided {obj} '{name}' "
                "has length {val}".format(
                    desired_shape=desired_shape, obj=object_name, name=name, val=values.shape
                )
            )
    else:
        values = np.asarray(values, dtype=np.float64)
        if values.shape != desired_shape:
            raise ValueError(
                "The desired shape of model state '{name}' is {desired_shape}, but provided {obj} '{name}' "
                "has length {val}".format(
                    desired_shape=desired_shape, obj=object_name, name=name, val=values.shape
                )
            )

    return values

def check_overlap(lst_1, lst_2, name_1, name_2):
    """A function raising an error if there is overlap between both lists"""
    overlap = set(lst_1) & set(lst_2)  # Find common elements between the lists
    if overlap:
        raise ValueError(f"overlap detected between '{name_1}' and '{name_2}': {list(overlap)}")

def check_duplicates(lst, name):
    """A function raising an error if lst contains duplicates"""
    seen = set()
    dupes = [x for x in lst if x in seen or seen.add(x)]
    if dupes:
        raise ValueError(
            f"List '{name}' contains duplicates: {dupes}"
        )

def merge_parameter_names_parameter_stratified_names(parameter_names, parameter_stratified_names):
    """ A function to merge the 'parameter_names' and 'parameter_stratified_names' lists"""
    merged_params = parameter_names.copy()
    if parameter_stratified_names:
        if not isinstance(parameter_stratified_names[0], list):
            if len(parameter_stratified_names) == 1:
                merged_params += parameter_stratified_names
            else:
                for stratified_names in parameter_stratified_names:
                    merged_params += [stratified_names,]
        else:
            for stratified_names in parameter_stratified_names:
                merged_params += stratified_names
    return merged_params

def validate_apply_transitionings_signature(apply_transitionings, parameter_names_merged, state_names):
    """A function checking the inputs of the apply_transitionings function"""

    # First argument should always be 't'
    sig = inspect.signature(apply_transitionings)
    keywords = list(sig.parameters.keys())
    if keywords[0] != "t":
        raise ValueError(
            "The first argument of the 'apply_transitionings' function should always be 't'"
        )
    elif keywords[1] != 'tau':
        raise ValueError(
            "The second argument of the 'apply_transitionings' function should always be 'tau'"
        )
    elif keywords[2] != 'transitionings':
        raise ValueError(
            "The third argument of the 'apply_transitionings' function should always be 'transitionings'"
        )

    # Verify all parameters and state follow after 't'/'tau'/'transitionings'
    if set(parameter_names_merged + state_names) != set(keywords[3:]):
        # Extract redundant and missing parameters
        redundant_all = set(keywords[3:]).difference(set(parameter_names_merged + state_names))
        missing_all = set(parameter_names_merged + state_names).difference(set(keywords[1:]))
        # Let's split the missing variables in parameters/states for extra clarity
        missing_states = [name for name in missing_all if name in state_names]
        missing_parameters = [name for name in missing_all if name in parameter_names_merged]
        raise ValueError(
            "The provided state names and parameters don't match the parameters and states of the apply_transitionings function. "
            "Missing parameters: {0}. Missing states: {1}. "
            "Redundant: {2}. ".format(missing_parameters, missing_states, list(redundant_all))
        )

def validate_integrate_or_compute_rates_signature(integrate_func, parameter_names_merged, state_names, _function_parameters):
    """A function checking the inputs of the integrate/compute_rates function"""

    # First argument should always be 't'
    sig = inspect.signature(integrate_func)
    keywords = list(sig.parameters.keys())
    if keywords[0] != "t":
        raise ValueError(
            "The first argument of the integrate/compute_rates function should always be 't'"
        )

    # Verify all parameters and state follow after 't'
    if set(parameter_names_merged + state_names) != set(keywords[1:]):
        # Extract redundant and missing parameters
        redundant_all = set(keywords[1:]).difference(set(parameter_names_merged + state_names))
        missing_all = set(parameter_names_merged + state_names).difference(set(keywords[1:]))
        # Let's split the missing variables in parameters/states for extra clarity
        missing_states = [name for name in missing_all if name in state_names]
        missing_parameters = [name for name in missing_all if name in parameter_names_merged]
        raise ValueError(
            "The input arguments of the integrate/compute_rates function should be 't', followed by the model's 'states' and 'parameters'. "
            "Missing parameters: {0}. Missing states: {1}. "
            "Redundant arguments: {2}. ".format(missing_parameters, missing_states, list(redundant_all))
        )

    all_params = parameter_names_merged.copy()
    # additional parameters from time-dependent parameter functions
    # are added to specified_params after the above check
    if _function_parameters:
        _extra_params = [item for sublist in _function_parameters for item in sublist]
        # Remove duplicate arguments in time dependent parameter functions
        _extra_params = OrderedDict((x, True) for x in _extra_params).keys()
        all_params += _extra_params
        len_before = len(all_params)
        # Line below removes duplicate arguments with integrate defenition
        all_params = list(OrderedDict((x, True) for x in all_params).keys())
        len_after = len(all_params)
        # Line below computes number of integrate arguments used in time dependent parameter functions
        n_duplicates = len_before - len_after
        _n_function_params = len(_extra_params) - n_duplicates
    else:
        _n_function_params = 0
        _extra_params = []
    
    return all_params, list(_extra_params)

def validate_provided_parameters(all_params, parameters, state_names):
    """
    Verify all parameters (parameters + stratified parameters + TDPF parameters + initial condition parameters) are provided by the user.
    Verify 't' is not used as a parameter.
    Verify no state name is used as a parameter
    """

    # Validate the type
    if not isinstance(parameters, dict):
        raise TypeError("'parameters' should be a dictionary.")
    
    # Validate the parameter dictionary provided by the user (does it contain missing/redundant parameters)
    if set(parameters.keys()) != set(all_params):
        raise ValueError(
            "The specified parameters don't exactly match the predefined parameters. "
            "Redundant parameters: {0}. Missing parameters: {1}".format(
            set(parameters.keys()).difference(set(all_params)),
            set(all_params).difference(set(parameters.keys())))
        )
    
    # Build a dictionary of parameters: key: parameter name, value: parameter value
    parameters = {param: parameters[param] for param in all_params}

    # After building the list of all model parameters, verify no parameters named 't' were used
    if 't' in parameters:
        raise ValueError(
        "'t' is not a valid model parameter name. please verify no model parameters named 't' are present in your model, its time-dependent parameter functions or initial condition function."
            )
    
    # Lastly, verify no state name was used a parameter
    overlap = set(state_names) & set(parameters.keys())  # Find common elements between the lists
    if overlap:
        raise ValueError(f"overlapping names of model parameters and states is not allowed. found overlapping names: {list(overlap)}")
    
    return parameters

def check_stratpar_size(values, name, object_name,dimension_names,desired_size):
    """Checks the size of stratified parameters. Converts stratified parameters to a numpy array."""
    values = np.asarray(values)
    if values.ndim != 1:
        raise ValueError(
            "A {obj} value should be a 1D array, but {obj} '{name}' is "
            "{val}-dimensional".format(
                obj=object_name, name=name, val=values.ndim
            )
        )
    if len(values) != desired_size:
        raise ValueError(
            "The coordinates provided for dimension '{dimension_names}' indicates a "
            "dimension size of {desired_size}, but {obj} '{name}' "
            "has length {val}".format(
                dimension_names=dimension_names, desired_size=desired_size,
                obj=object_name, name=name, val=len(values)
            )
        )
    return values

def validate_parameter_stratified_sizes(parameter_stratified_names, dimension_names, coordinates, parameters):
    """Check if the sizes of the stratified parameters are correct"""

    if not isinstance(parameter_stratified_names[0], list):
        if len(list(coordinates.keys())) > 1:
            raise ValueError(
                f"The model has more than one dimension ({len(list(coordinates.keys()))}). I cannot deduce the dimension of your statified parameter from the provided `parameter_stratified_names`: {parameter_stratified_names}. "
                f"Make sure `parameter_stratified_names` is a list, containing {len(list(coordinates.keys()))} (no. dimensions) sublists. Each sublist corresponds to a dimension in `dimension_names`. "
                "Place your stratified parameter in the correct sublist so I know what model dimension to match it with. If a dimension has no stratified parameter, provide an empty list."
            )
        else:
            for param in parameter_stratified_names:
                parameters[param] = check_stratpar_size(
                                        parameters[param], param, "stratified parameter",dimension_names[0],len(coordinates[dimension_names[0]])
                                        )
    else:
        # Number of sublists equal to number of dimensions?
        if len(parameter_stratified_names) != len(list(coordinates.keys())):
            raise ValueError(
                f"The number of sublists in `parameter_stratified_names` ({len(parameter_stratified_names)}) must be equal to the number of coordinates ({len(list(coordinates.keys()))})"
            )
        for j,stratified_names in enumerate(parameter_stratified_names):
            for param in stratified_names:
                parameters[param] = check_stratpar_size(
                                        parameters[param], param, "stratified parameter",dimension_names[j],len(coordinates[dimension_names[j]])
                                        )
    return parameters

def validate_integrate(initial_states, parameters, integrate, state_shapes):
    """ Call _create_fun to check if the integrate function (1) works, (2) the differentials have the correct shape
    """

    try:
        # Call the integrate function
        dstates = integrate(1, **initial_states, **parameters)
        # Flatten
        out=[]
        for d in dstates:
            out.extend(list(np.ravel(d)))
    except:
        raise ValueError(
            "An error was encountered while calling your integrate function."
        )
    # Flatten initial states
    y0=[]
    for v in initial_states.values():
        y0.extend(list(np.ravel(v)))
    # Assert length equality
    if len(out) != len(y0):
        raise ValueError(
            "The total length of the differentials returned by your `integrate()` function do not appear to have the correct length. "
            "Verify the differentials are in the same order as `state_names`. Verify every differential has the same size as the state it corresponds to. "
            f"State shapes: {state_shapes}"
        )

def validate_compute_rates(compute_rates, initial_states, states_shape, parameter_names_merged, parameters):
    """A function to call compute_rates and check if the output type and contents are alright"""

    # Throw TDPF parameters out of parameters dictionary
    parameters_wo_TDPF_pars={k:v for k, v in parameters.items() if k in parameter_names_merged}
    # 'func' in class 'JumpProcess' of 'base.py' automatically converts states to np.array
    # However, we do not wish to validate the output of 'func' but rather of its consituent functions: compute_rates, apply_transitionings
    initial_states_copy={k: v[:] for k, v in initial_states.items()}
    for k,v in initial_states.items():
        initial_states_copy[k] = np.asarray(initial_states[k])
    # Call the function with initial values to check if the function returns the right format of dictionary
    rates = compute_rates(10, **initial_states_copy, **parameters_wo_TDPF_pars)
    # Check if a dictionary is returned
    if not isinstance(rates, dict):
        raise TypeError("Output of function 'compute_rates' should be of type dictionary")
    # Check if all states present are valid model parameters
    for state in rates.keys():
        if not state in initial_states.keys():
            raise ValueError(
                f"Rate found for an invalid model state '{state}'"
            )
    # Check if all rates are given as a list
    for state_name, rate_list in rates.items():
        if not isinstance(rate_list, list):
            raise ValueError(
                f"Rate(s) of model state '{state_name}' are not containted within a list."
            )
    # Check if the sizes are correct
    for state_name, rate_list in rates.items():
        for i,rate in enumerate(rate_list):
            if not isinstance(rate, np.ndarray):
                raise TypeError(f"the rate of the {i}th transitioning for state '{state_name}' coming from function `compute_rates` is not of type 'np.ndarray' but {type(rate)}")
            if rate.shape != states_shape[state_name]:
                raise ValueError(f"State {state_name} has size {states_shape[state_name]}, but rate of the {i}th transitioning for state '{state_name}' has shape {list(rate.shape)}")
    return rates

def validate_apply_transitionings(apply_transitionings, rates, initial_states, states_shape, parameter_names_merged, parameters):
    """A function to call apply_transitionings and check if the output type and contents are alright"""
    
    # Throw TDPF parameters out of parameters dictionary
    parameters_wo_TDPF_pars={k:v for k, v in parameters.items() if k in parameter_names_merged}
    # Update states
    new_states = apply_transitionings(10, 1, rates, **initial_states, **parameters_wo_TDPF_pars)
    # Check
    if len(list(new_states)) != len(initial_states.keys()):
        raise ValueError(f"The number of outputs of function 'apply_transitionings_func' ({len(list(new_states))}) is not equal to the number of states ({len(initial_states.keys())})")
    for i, new_state in enumerate(list(new_states)):
        if not isinstance(new_state, np.ndarray):
            raise TypeError(f"Output state of function 'apply_transitionings_func' in position {i} is not a np.ndarray")
        if new_state.shape != states_shape[list(initial_states.keys())[i]]:
                raise ValueError(f"State {list(initial_states.keys())[i]} has size {states_shape[list(initial_states.keys())[i]]}, but rate of the {i}th transitioning for state '{list(initial_states.keys())[i]}' has shape {list(rate.shape)}")


from setuptools import find_packages, setup

setup(
    name='pySODM',
    packages=find_packages("src", exclude=["*.tests"]),
    package_dir={'': 'src'},
    version='0.2.8',
    description='Simulating and Optimising Dynamical Models',
    author='Tijs Alleman (Ghent University, Cornell University)',
    author_email='tijs.alleman@ugent.be, twa27@cornell.edu',
    keywords='ODE PDE simulation calibration gillespie xarray emcee',
    url='https://github.com/twallema/pySODM',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license='CC-BY-NC-SA',
    python_requires='>3.11.11',
    install_requires=[
        'scipy',
        'numpy',
        'pandas',
        'matplotlib',
        'xarray',
        'emcee',
        'h5py',
        'tqdm'
    ],
    extras_require={
        "develop":  ["pytest",
                     "sphinx",
                     "setuptools",
                     "numpydoc",
                     "sphinx_rtd_theme",
                     "myst_parser[sphinx]"],
    }
)